# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019-2020 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging

from aioxmpp import JID
from typing import TYPE_CHECKING, Any, List, Optional
from .enums import UserSearchPlatform, UserSearchMatchType, StatsCollectionType
from .typedefs import DatetimeOrTimestamp
from .errors import Forbidden

if TYPE_CHECKING:
    from .client import Client
    from .stats import StatsV2, StatsCollection

log = logging.getLogger(__name__)


class ExternalAuth:
    """Represents an external auth belonging to a user.

    Attributes
    ----------
    client: :class:`Client`
        The client.
    type: :class:`str`:
        The type/platform of the external auth.
    id: :class:`str`:
        The users universal fortnite id.
    external_id: :class:`str`:
        The id belonging to this user on the platform.
    external_display_name: :class:`str`
        The display name belonging to this user on the platform.
    extra_info: Dict[:class:`str`, Any]
        Extra info from the payload. Usually empty on accounts other
        than :class:`ClientUser`.
    """

    __slots__ = ('client', 'type', 'id', 'external_id',
                 'external_display_name', 'extra_info')

    def __init__(self, client: 'Client', data: dict) -> None:
        self.client = client
        self.type = data['type']
        self.id = data['accountId']
        self.external_id = data['externalAuthId']
        self.external_display_name = data['externalDisplayName']

    def _update_extra_info(self, data: dict) -> None:
        to_be_removed = ('type', 'accountId', 'externalAuthId',
                         'externalDisplayName')
        for field in to_be_removed:
            try:
                del data[field]
            except KeyError:
                pass

        self.extra_info = data

    def __str__(self) -> str:
        return self.external_display_name

    def __repr__(self) -> str:
        return ('<ExternalAuth type={0.type!r} id={0.id!r} '
                'external_display_name={0.external_display_name!r} '
                'external_id={0.external_id!r}>'.format(self))

    def __eq__(self, other):
        return isinstance(other, ExternalAuth) and other.id == self.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def get_raw(self) -> dict:
        return {
            'type': self.type,
            'accountId': self.id,
            'externalAuthId': self.external_id,
            'externalDisplayName': self.external_display_name,
            **self.extra_info
        }


class UserBase:
    __slots__ = ('client', '_epicgames_display_name', '_external_display_name',
                 '_id', '_external_auths')

    def __init__(self, client: 'Client', data: dict, **kwargs: Any) -> None:
        self.client = client
        if data:
            self._update(data)

    def __str__(self) -> str:
        return self.display_name

    def __eq__(self, other):
        return isinstance(other, UserBase) and other._id == self._id

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def display_name(self) -> str:
        """:class:`str`: The users displayname

        .. warning::

            The display name will be the one registered to the epicgames
            account. If an epicgames account is not found it defaults
            to the display name of an external auth.
        """
        return self._epicgames_display_name or self._external_display_name

    @property
    def id(self) -> str:
        """:class:`str`: The users id"""
        return self._id

    @property
    def external_auths(self) -> List[ExternalAuth]:
        """List[:class:`ExternalAuth`]: List containing information about
        external auths. Might be empty if the user does not have any external
        auths.
        """
        return self._external_auths

    @property
    def epicgames_account(self) -> bool:
        """:class:`bool`: Tells you if the user is an account registered to epicgames
        services. ``False`` if the user is from another platform without
        having linked their account to an epicgames account.

        .. warning::

            If this is True, the display name will be the one registered to
            the epicgames account, if not it defaults to the display name of
            an external auth.
        """
        return self._epicgames_display_name is not None

    @property
    def jid(self) -> JID:
        """:class:`aioxmpp.JID`: The JID of the user."""
        return JID.fromstr('{0.id}@{0.client.service_host}'.format(self))

    async def fetch_br_stats(self, *,
                             start_time: Optional[DatetimeOrTimestamp] = None,
                             end_time: Optional[DatetimeOrTimestamp] = None
                             ) -> 'StatsV2':
        """|coro|

        Fetches this users stats.

        Parameters
        ----------
        start_time: Optional[Union[:class:`int`, :class:`datetime.datetime`, :class:`SeasonStartTimestamp`]]
            The UTC start time of the time period to get stats from.
            *Must be seconds since epoch, :class:`datetime.datetime` or a constant from SeasonEndTimestamp*
            *Defaults to None*
        end_time: Optional[Union[:class:`int`, :class:`datetime.datetime`, :class:`SeasonEndTimestamp`]]
            The UTC end time of the time period to get stats from.
            *Must be seconds since epoch, :class:`datetime.datetime` or a constant from SeasonEndTimestamp*
            *Defaults to None*

        Raises
        ------
        Forbidden
            The user has chosen to be hidden from public stats by disabling
            the fortnite setting below.
            ``Settings`` -> ``Account and Privacy`` -> ``Show on career
            leaderboard``
        HTTPException
            An error occured while requesting.

        Returns
        -------
        :class:`StatsV2`
            An object representing the stats for this user.
        """  # noqa
        return await self.client.fetch_br_stats(
            self.id,
            start_time=start_time,
            end_time=end_time
        )

    async def fetch_br_stats_collection(self, collection: StatsCollectionType,
                                        start_time: Optional[DatetimeOrTimestamp] = None,  # noqa
                                        end_time: Optional[DatetimeOrTimestamp] = None  # noqa)
                                        ) -> 'StatsCollection':
        """|coro|

        Fetches a stats collections for this user.

        Parameters
        ----------
        start_time: Optional[Union[:class:`int`, :class:`datetime.datetime`, :class:`SeasonStartTimestamp`]]
            The UTC start time of the time period to get stats from.
            *Must be seconds since epoch, :class:`datetime.datetime` or a constant from SeasonEndTimestamp*
            *Defaults to None*
        end_time: Optional[Union[:class:`int`, :class:`datetime.datetime`, :class:`SeasonEndTimestamp`]]
            The UTC end time of the time period to get stats from.
            *Must be seconds since epoch, :class:`datetime.datetime` or a constant from SeasonEndTimestamp*
            *Defaults to None*

        Raises
        ------
        Forbidden
            The user has chosen to be hidden from public stats by disabling
            the fortnite setting below.
            ``Settings`` -> ``Account and Privacy`` -> ``Show on career
            leaderboard``
        HTTPException
            An error occured while requesting.

        Returns
        -------
        :class:`StatsCollection`
            An object representing the stats collection for this user.
        """  # noqa
        res = await self.client.fetch_multiple_br_stats_collections(
            user_ids=(self.id,),
            collection=collection,
            start_time=start_time,
            end_time=end_time,
        )

        if self.id not in res:
            raise Forbidden('User has opted out of public leaderboards.')

        return res[self.id]

    async def fetch_battlepass_level(self, *,
                                     season: int,
                                     start_time: Optional[DatetimeOrTimestamp] = None,  # noqa
                                     end_time: Optional[DatetimeOrTimestamp] = None  # noqa
                                     ) -> float:
        """|coro|

        Fetches this users battlepass level.

        Parameters
        ----------
        season: :class:`int`
            The season number to request the battlepass level for.
        start_time: Optional[Union[:class:`int`, :class:`datetime.datetime`, :class:`SeasonStartTimestamp`]]
            The UTC start time of the window to get the battlepass level from.
            *Must be seconds since epoch, :class:`datetime.datetime` or a constant from SeasonEndTimestamp*
            *Defaults to None*
        end_time: Optional[Union[:class:`int`, :class:`datetime.datetime`, :class:`SeasonEndTimestamp`]]
            The UTC end time of the window to get the battlepass level from.
            *Must be seconds since epoch, :class:`datetime.datetime` or a constant from SeasonEndTimestamp*
            *Defaults to None*

        Raises
        ------
        HTTPException
            An error occured while requesting.

        Returns
        -------
        Optional[:class:`float`]
            The users battlepass level. ``None`` is returned if the user has
            not played any real matches this season.

            .. note::

                The decimals are the percent progress to the next level.
                E.g. ``208.63`` -> ``Level 208 and 63% on the way to 209.``
        """  # noqa
        return await self.client.fetch_battlepass_level(
            self.id,
            season=season,
            start_time=start_time,
            end_time=end_time
        )

    def _update(self, data: dict) -> None:
        self._epicgames_display_name = data.get('displayName',
                                                data.get('account_dn'))
        self._update_external_auths(
            data.get('externalAuths', data.get('external_auths', [])),
            extra_external_auths=data.get('extraExternalAuths', []),
        )

        self._id = data.get('accountId',
                            data.get('id', data.get('account_id')))

    def _update_external_auths(self, external_auths: List[dict], *,
                               extra_external_auths: List[dict] = []) -> None:
        extra_ext = {v['authIds'][0]['type'].split('_')[0].lower(): v
                     for v in extra_external_auths}

        ext_list = []
        for e in external_auths:
            ext = ExternalAuth(self.client, e)
            ext._update_extra_info(extra_ext.get(ext.type, {}))
            ext_list.append(ext)

        self._external_display_name = None
        for ext_auth in reversed([x for x in ext_list
                                  if x.type.lower() not in ('twitch',)]):
            self._external_display_name = ext_auth.external_display_name
            break

        self._external_auths = ext_list

    def _update_epicgames_display_name(self, display_name: str) -> None:
        self._epicgames_display_name = display_name

    def get_raw(self) -> dict:
        return {
            'displayName': self.display_name,
            'id': self.id,
            'externalAuths': [ext.get_raw() for ext in self._external_auths]
        }


class ClientUser(UserBase):
    """Represents the user the client is connected to.

    Attributes
    ----------
    client: :class:`Client`
        The client.
    age_group: :class:`str`
        The age group of the user.
    can_update_display_name: :class:`bool`
        ``True`` if the user can update it's displayname else ``False``
    country: :class:`str`
        The country the user wasregistered in.
    email: :class:`str`
        The email of the user.
    failed_login_attempts: :class:`str`
        Failed login attempts
    headless: :class:`bool`
        ``True`` if the account has no display name due to no epicgames
        account being linked to the current account.
    last_login: :class:`datetime.datetime`
        UTC time of the last login of the user. ``None`` if no failed login
        attempt has been registered.
    name: :class:`str`
        First name of the user.
    first_name: :class:`str`
        First name of the user. Alias for name.
    last_name: :class:`str`
        Last name of the user.
    full_name: :class:`str`
        Full name of the user.
    number_of_display_name_changes: :class:`int`
        Amount of displayname changes.
    preferred_language: :class:`str`
        Users preferred language.
    tfa_enabled: :class:`bool`
        ``True`` if the user has two-factor authentication enabled else
        ``False``.
    email_verified: :class:`bool`
        ``True`` if the accounts email has been verified.
    minor_verified: :class:`bool`
        ``True`` if the account has been verified to be run by a minor.
    minor_expected: :class:`bool`
        ``True`` if the account is expected to be run by a minor.
    minor_status: :class:`str`
        The minor status of this account.
    """

    def __init__(self, client: 'Client', data: dict, **kwargs: Any) -> None:
        super().__init__(client, data)
        self._update(data)

    def __repr__(self) -> str:
        return ('<ClientUser id={0.id!r} display_name={0.display_name!r} '
                'jid={0.jid!r} email={0.email!r}>'.format(self))

    @property
    def first_name(self) -> str:
        return self.name

    @property
    def full_name(self) -> str:
        return '{} {}'.format(self.name, self.last_name)

    @property
    def jid(self) -> JID:
        """:class:`aioxmpp.JID`: The JID of the client. Includes the
        resource part.
        """
        return self.client.xmpp.xmpp_client.local_jid

    def _update(self, data: dict) -> None:
        super()._update(data)
        self.name = data['name']
        self.email = data['email']
        self.failed_login_attempts = data['failedLoginAttempts']
        self.last_failed_login = (self.client.from_iso(data['lastFailedLogin'])
                                  if 'lastFailedLogin' in data else None)
        self.last_login = (self.client.from_iso(data['lastLogin'])
                           if 'lastLogin' in data else None)

        n_changes = data['numberOfDisplayNameChanges']
        self.number_of_display_name_changes = n_changes
        self.age_group = data['ageGroup']
        self.headless = data['headless']
        self.country = data['country']
        self.last_name = data['lastName']
        self.preferred_language = data['preferredLanguage']
        self.can_update_display_name = data['canUpdateDisplayName']
        self.tfa_enabled = data['tfaEnabled']
        self.email_verified = data['emailVerified']
        self.minor_verified = data['minorVerified']
        self.minor_expected = data['minorExpected']
        self.minor_status = data['minorStatus']


class User(UserBase):
    """Represents a user from Fortnite"""

    __slots__ = UserBase.__slots__

    def __init__(self, client: 'Client', data: dict, **kwargs: Any) -> None:
        super().__init__(client, data)

    def __repr__(self) -> str:
        return ('<User id={0.id!r} display_name={0.display_name!r} '
                'epicgames_account={0.epicgames_account!r}>'.format(self))

    async def block(self) -> None:
        """|coro|

        Blocks this user.

        Raises
        ------
        HTTPException
            Something went wrong while blocking this user.
        """
        await self.client.block_user(self.id)

    async def add(self) -> None:
        """|coro|

        Sends a friendship request to this user or adds them if they
        have already sent one to the client.

        Raises
        ------
        NotFound
            The specified user does not exist.
        DuplicateFriendship
            The client is already friends with this user.
        FriendshipRequestAlreadySent
            The client has already sent a friendship request that has not been
            handled yet by the user.
        MaxFriendshipsExceeded
            The client has hit the max amount of friendships a user can
            have at a time. For most accounts this limit is set to ``1000``
            but it could be higher for others.
        InviteeMaxFriendshipsExceeded
            The user you attempted to add has hit the max amount of friendships
            a user can have at a time.
        InviteeMaxFriendshipRequestsExceeded
            The user you attempted to add has hit the max amount of friendship
            requests a user can have at a time. This is usually ``700`` total
            requests.
        Forbidden
            The client is not allowed to send friendship requests to the user
            because of the users settings.
        HTTPException
            An error occured while requesting to add this friend.
        """
        await self.client.add_friend(self.id)


class BlockedUser(UserBase):
    """Represents a blocked user from Fortnite"""

    __slots__ = UserBase.__slots__

    def __init__(self, client: 'Client', data: dict) -> None:
        super().__init__(client, data)

    def __repr__(self) -> str:
        return ('<BlockedUser id={0.id!r} '
                'display_name={0.display_name!r} '
                'epicgames_account={0.epicgames_account!r}>'.format(self))

    async def unblock(self) -> None:
        """|coro|

        Unblocks this friend.
        """
        await self.client.unblock_user(self.id)


class UserSearchEntry(User):
    """Represents a user entry in a user search.

    Parameters
    ----------
    matches: List[Tuple[:class:`str`, :class:`UserSearchPlatform`]]
        | A list of tuples containing the display name the user matched
        and the platform the display name is from.
        | Example: ``[('Tfue', UserSearchPlatform.EPIC_GAMES)]``
    match_type: :class:`UserSearchMatchType`
        The type of match this user matched by.
    mutual_friend_count: :class:`int`
        The amount of **epic** mutual friends the client has with the user.
    """
    def __init__(self, client: 'Client',
                 user_data: dict,
                 search_data: dict) -> None:
        super().__init__(client, user_data)

        self.matches = [(d['value'], UserSearchPlatform(d['platform']))
                        for d in search_data['matches']]
        self.match_type = UserSearchMatchType(search_data['matchType'])
        self.mutual_friend_count = search_data['epicMutuals']

    def __str__(self) -> str:
        return self.matches[0][0]

    def __repr__(self) -> str:
        return ('<UserSearchEntry id={0.id!r} '
                'display_name={0.display_name!r} '
                'epicgames_account={0.epicgames_account!r}>'.format(self))


class SacSearchEntryUser(User):
    """Represents a user entry in a support a creator code search.

    Parameters
    ----------
    slug: :class:`str`
        The slug (creator code) that matched.
    active: :class:`bool`
        Wether or not the creator code is active or not.
    verified: :class:`bool`
        Wether or not the creator code is verified or not.
    """
    def __init__(self, client: 'Client',
                 user_data: dict,
                 search_data: dict) -> None:
        super().__init__(client, user_data)

        self.slug = search_data['slug']
        self.active = search_data['status'] == 'ACTIVE'
        self.verified = search_data['verified']

    def __repr__(self) -> str:
        return ('<SacSearchEntryUser slug={0.slug!r} '
                'id={0.id!r} '
                'display_name={0.display_name!r} '
                'epicgames_account={0.epicgames_account!r}>'.format(self))



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019-2020 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import aioxmpp
import asyncio
import json
import logging
import datetime
import uuid
import itertools
import unicodedata
import aiohttp

from xml.etree import ElementTree
from collections import defaultdict
from typing import TYPE_CHECKING, Optional, Union, Awaitable, Any, Tuple

from .errors import XMPPError, PartyError, HTTPException
from .message import FriendMessage, PartyMessage
from .party import Party, ReceivedPartyInvitation, PartyJoinConfirmation
from .presence import Presence
from .enums import AwayStatus

if TYPE_CHECKING:
    from .client import Client

log = logging.getLogger(__name__)


def is_RandALCat(c: str) -> bool:
    return unicodedata.bidirectional(c) in ('R', 'AL')


class EventContext:

    __slots__ = ('client', 'body', 'party', 'created_at')

    def __init__(self, client: 'Client', body: dict) -> None:
        self.client = client
        self.body = body

        self.party = self.client.party
        self.created_at = datetime.datetime.utcnow()


class EventDispatcher:
    def __init__(self) -> None:
        self._listeners = defaultdict(list)
        self._presence_listeners = []
        self.interactions_enabled = False

    def process_presence(self, client, *args) -> None:
        for coro in self._presence_listeners:
            if __name__ == coro.__module__:
                asyncio.ensure_future(coro(client.xmpp, *args))
            else:
                asyncio.ensure_future(coro(*args))

    def presence(self) -> Awaitable:
        def decorator(coro: Awaitable) -> Awaitable:
            self.add_presence_handler(coro)
            return coro
        return decorator

    def add_presence_handler(self, coro: Awaitable) -> None:
        if coro not in self._presence_listeners:
            self._presence_listeners.append(coro)

    def remove_presence_handler(self, coro: Awaitable) -> None:
        self._presence_listeners = [
            c for c in self._presence_listeners if c is not coro
        ]

    def process_event(self, client: 'Client', raw_body: dict) -> None:
        body = json.loads(raw_body)

        type_ = body.get('type')
        if type_ is None:
            if self.interactions_enabled:
                for interaction in body['interactions']:
                    self.process_event(client, interaction)
            return

        log.debug('Received event `{}` with body `{}`'.format(type_, body))

        coros = self._listeners.get(type_, [])
        for coro in coros:
            ctx = EventContext(client, body)

            if __name__ == coro.__module__:
                asyncio.ensure_future(coro(client.xmpp, ctx))
            else:
                asyncio.ensure_future(coro(ctx))

    def event(self, event: str) -> Awaitable:
        def decorator(coro: Awaitable) -> Awaitable:
            self.add_event_handler(event, coro)
            return coro
        return decorator

    def add_event_handler(self, event: str, coro: Awaitable) -> None:
        self._listeners[event].append(coro)
        log.debug('Added handler for {0} to {1}'.format(event, coro))

    def remove_event_handler(self, event: str, coro: Awaitable) -> None:
        handlers = [c for c in self._listeners[event] if c is not coro]
        log.debug('Removed {0} handler(s) for {1}'.format(
            len(self._listeners[event]) - len(handlers),
            event
        ))
        self._listeners[event] = handlers


dispatcher = EventDispatcher()


class XMLProcessor:
    def _process_presence(self, raw: str) -> Optional[Union[tuple, bool]]:
        tree = ElementTree.fromstring(raw)

        type_ = tree.get('type')

        # Only intercept presences with either no type attribute
        # (which means available) or unavailable type.
        if type_ is not None and type_ not in ('available', 'unavailable'):
            return False

        from_ = tree.get('from')

        # If from is a party, let aioxmpp handle it.
        if from_ is not None and '-' in from_:
            return False

        status = None
        show = None
        for elem in tree:
            if 'status' in elem.tag:
                status = elem.text
            if 'show' in elem.tag:
                show = elem.text

        # We have no use for the presence if status is None and
        # therefore it's better to just let aioxmpp handle it.
        if status is None:
            return False

        split = from_.split('@')
        user_id = split[0]
        platform = split[1].split(':')[2]

        return 'presence', (user_id, platform, type_, status, show)

    def _process_message(self, raw: str) -> Optional[Union[tuple, bool]]:
        tree = ElementTree.fromstring(raw)

        type_ = tree.get('type')

        # Only intercept messages with either no type attribute
        # (which means normal) or  type.
        if type_ is not None and type_ != 'normal':
            return False

        # Let aioxmpp handle it if no body tag is found. Also, technically
        # a message can include multiple body tags for different languages
        # but afaik only one body tag is sent from epics servers.
        body = None
        for elem in tree:
            if 'body' in elem.tag:
                body = elem.text
                break
        else:
            return False

        return 'message', (body,)

    def process(self, raw: str) -> Optional[Union[tuple, bool]]:
        # Yes, this is a hacky solution but it's better than
        # using the quite unnecessary slow aioxmpp one.
        if '<presence' in raw:
            return self._process_presence(raw)
        elif '<message' in raw:
            return self._process_message(raw)

        return False


class WebsocketTransport:
    def __init__(self, loop: asyncio.AbstractEventLoop,
                 stream: 'WebsocketXMLStream',
                 client: 'Client',
                 logger: logging.Logger,
                 ws_connector: Optional[aiohttp.BaseConnector] = None) -> None:
        self.loop = loop
        self.stream = stream
        self.client = client
        self.logger = logger
        self.ws_connector = ws_connector

        self.xml_processor = XMLProcessor()

        self.connection = None
        self._buffer = b''
        self._reader_task = None
        self._close_event = asyncio.Event()
        self._called_lost = False
        self._attempt_reconnect = True

    async def create_connection(self, *args,
                                **kwargs) -> aiohttp.ClientWebSocketResponse:
        self.logger.debug('Setting up new websocket connection.')

        self.session = aiohttp.ClientSession(
            connector=self.ws_connector,
            connector_owner=self.ws_connector is None,
        )
        self.connection = con = await self.session.ws_connect(
            *args, **kwargs
        )

        self.loop.create_task(self.reader())
        self.stream.connection_made(self)
        self._called_lost = False
        self._attempt_reconnect = True

        self.logger.debug('Websocket connection established.')
        return con

    async def reader(self) -> None:
        self.logger.debug('Websocket reader is now running.')

        try:
            while True:
                msg = await self.connection.receive()

                self.logger.debug('RECV: {0}'.format(msg))
                if msg.type == aiohttp.WSMsgType.TEXT:
                    ret = self.xml_processor.process(msg.data)
                    if ret is None:
                        continue
                    elif ret is False:
                        self.stream.data_received(msg.data)
                    else:
                        type_ = ret[0]
                        if type_ == 'presence':
                            dispatcher.process_presence(
                                self.client,
                                *ret[1]
                            )
                        elif type_ == 'message':
                            dispatcher.process_event(
                                self.client,
                                *ret[1]
                            )

                if msg.type == aiohttp.WSMsgType.CLOSED:
                    if self._attempt_reconnect:
                        err = ConnectionError(
                            'websocket stream closed by peer'
                        )
                    else:
                        err = None

                    if not self._called_lost:
                        self._called_lost = True
                        self.stream.connection_lost(err)
                    break

                if msg.type == aiohttp.WSMsgType.ERROR:
                    if not self._called_lost:
                        self._called_lost = True
                        self.stream.connection_lost(
                            ConnectionError(
                                'websocket stream received an error: '
                                '{0}'.format(self.connection.exception()))
                        )
                    break
        finally:
            self.logger.debug('Websocket reader stopped.')

    async def send(self, data: bytes) -> None:
        self.logger.debug('SEND: {0}'.format(data))
        await self.connection.send_bytes(data)

    def write(self, data: bytes) -> None:
        self._buffer += data

    def flush(self) -> None:
        if self._buffer:
            asyncio.ensure_future(self.send(self._buffer))

        self._buffer = b''

    def can_write_eof(self) -> bool:
        return False

    def write_eof(self) -> None:
        raise NotImplementedError("Cannot write_eof() on ws transport.")

    def _stop_reader(self) -> None:
        if self._reader_task is not None and not self._reader_task.cancelled():
            self._reader_task.cancel()

    def close_callback(self, *args) -> None:
        self._close_event.set()

    def on_close(self, *args) -> None:
        try:
            task = self.loop.create_task(self.session.close())
        except AttributeError:
            pass
        else:
            task.add_done_callback(self.close_callback)

    def _close(self) -> None:
        if not self.connection:
            raise RuntimeError('Cannot close a non-existing connection.')

        self.logger.debug('Closing websocket connection.')

        task = self.loop.create_task(self.connection.close())
        task.add_done_callback(self.on_close)

        self._stop_reader()

    def close(self) -> None:
        self._attempt_reconnect = False
        self._close()

    def abort(self) -> None:
        self.logger.debug('Received abort signal.')
        self._close()

    async def wait_closed(self) -> None:
        await self._close_event

    def get_extra_info(self, *args, **kwargs) -> Any:
        return self.connection.get_extra_info(*args, **kwargs)


class WebsocketXMLStreamWriter(aioxmpp.xml.XMLStreamWriter):
    def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        self._writer.endElementNS(
            (aioxmpp.utils.namespaces.xmlstream, "stream"),
            None
        )
        for prefix in self._nsmap_to_use:
            self._writer.endPrefixMapping(prefix)
        self._writer.endDocument()
        self._writer.flush()
        del self._writer


class WebsocketXMLStream(aioxmpp.protocol.XMLStream):
    def _reset_state(self) -> None:
        self._kill_state()

        self._processor = aioxmpp.xml.XMPPXMLProcessor()
        self._processor.stanza_parser = self.stanza_parser
        self._processor.on_stream_header = self._rx_stream_header
        self._processor.on_stream_footer = self._rx_stream_footer
        self._processor.on_exception = self._rx_exception
        self._parser = aioxmpp.xml.make_parser()
        self._parser.setContentHandler(self._processor)
        self._debug_wrapper = None

        if self._logger.getEffectiveLevel() <= logging.DEBUG:
            dest = aioxmpp.protocol.DebugWrapper(self._transport, self._logger)
            self._debug_wrapper = dest
        else:
            dest = self._transport

        self._writer = WebsocketXMLStreamWriter(
            dest,
            self._to,
            nsmap={None: "jabber:client"},
            sorted_attributes=self._sorted_attributes)


class XMPPOverWebsocketConnector(aioxmpp.connector.BaseConnector):
    def __init__(self, client, ws_connector=None):
        self.client = client
        self.ws_connector = ws_connector

    @property
    def tls_supported(self) -> bool:
        return False

    @property
    def dane_supported(self) -> bool:
        return False

    async def connect(self, loop: asyncio.AbstractEventLoop,
                      metadata: aioxmpp.security_layer.SecurityLayer,
                      domain: str,
                      host: str,
                      port: int,
                      negotiation_timeout: Union[int, float],
                      base_logger: Optional[logging.Logger] = None
                      ) -> Tuple[WebsocketTransport,
                                 WebsocketXMLStream,
                                 aioxmpp.nonza.StreamFeatures]:
        features_future = asyncio.Future()

        stream = WebsocketXMLStream(
            to=domain,
            features_future=features_future,
            base_logger=base_logger,
        )

        if base_logger is not None:
            logger = base_logger.getChild(type(self).__name__)
        else:
            logger = logging.getLogger(".".join([
                __name__, type(self).__qualname__,
            ]))

        transport = WebsocketTransport(
            loop,
            stream,
            self.client,
            logger,
            ws_connector=self.ws_connector
        )
        await transport.create_connection(
            'wss://{host}'.format(host=host),
            protocols=('xmpp',),
            timeout=10,
        )

        return transport, stream, await features_future


# Were just patching this method to suppress an exception
# which is raised on stream error.
def _patched_done_handler(self, task):
    try:
        task.result()
    except asyncio.CancelledError:
        pass
    except Exception as err:
        try:
            if self._sm_enabled:
                self._xmlstream.abort()
            else:
                self._xmlstream.close()
        except Exception:
            pass
        self.on_failure(err)


aioxmpp.stream.StanzaStream._done_handler = _patched_done_handler


class XMPPClient:
    def __init__(self, client: 'Client', ws_connector=None) -> None:
        self.client = client
        self.ws_connector = ws_connector

        self.xmpp_client = None
        self.stream = None
        self.muc_room = None

        self._ping_task = None
        self._is_suspended = False
        self._reconnect_recover_task = None
        self._last_disconnected_at = None
        self._last_known_party_id = None
        self._task = None

    def jid(self, user_id: str) -> aioxmpp.JID:
        return aioxmpp.JID.fromstr('{}@{}'.format(
            user_id,
            self.client.service_host
        ))

    def _remove_illegal_characters(self, chars: str) -> str:
        fixed = []
        for c in chars:
            try:
                aioxmpp.stringprep.resourceprep(c)
            except ValueError:
                continue

            if is_RandALCat(c):
                continue

            fixed.append(c)
        return ''.join(fixed)

    def _create_invite(self, from_id: str, data: dict) -> dict:
        sent_at = self.client.from_iso(data['sent'])
        expires_at = sent_at + datetime.timedelta(hours=4)

        for m in data['members']:
            if m['account_id'] == from_id:
                member = m
                break

        party_m = data['meta']
        member_m = member['meta']

        meta = {
            'urn:epic:conn:type_s': 'game',
            'urn:epic:cfg:build-id_s': party_m['urn:epic:cfg:build-id_s'],
            'urn:epic:invite:platformdata_s': '',
        }

        if 'Platform_j' in member_m:
            meta['Platform_j'] = json.loads(
                member_m['Platform_j']
            )['Platform']['platformStr']

        if 'urn:epic:member:dn_s' in member['meta']:
            meta['urn:epic:member:dn_s'] = member_m['urn:epic:member:dn_s']

        inv = {
            'party_id': data['id'],
            'sent_by': from_id,
            'sent_to': self.client.user.id,
            'sent_at': self.client.to_iso(sent_at),
            'updated_at': self.client.to_iso(sent_at),
            'expires_at': self.client.to_iso(expires_at),
            'status': 'SENT',
            'meta': meta
        }
        return inv

    async def process_chat_message(self, message: aioxmpp.Message) -> None:
        user_id = message.from_.localpart
        author = self.client.get_friend(user_id)
        if author is None:
            try:
                author = await self.client.wait_for(
                    'friend_add',
                    check=lambda f: f.id == user_id,
                    timeout=2
                )
            except asyncio.TimeoutError:
                log.debug('Friend message discarded because friend not found.')
                return

        try:
            m = FriendMessage(
                client=self.client,
                author=author,
                content=message.body.any()
            )
            self.client.dispatch_event('friend_message', m)
        except ValueError:
            pass

    @dispatcher.event('com.epicgames.friends.core.apiobjects.Friend')
    async def friend_event(self, ctx: EventContext) -> None:
        body = ctx.body

        await self.client.wait_until_ready()
        _payload = body['payload']
        _status = _payload['status']
        _id = _payload['accountId']

        if _status == 'ACCEPTED':

            data = self.client.get_user(_id)
            if data is None:
                data = await self.client.fetch_user(_id, raw=True)
            else:
                data = data.get_raw()

            try:
                timestamp = body['timestamp']
            except (TypeError, KeyError):
                timestamp = datetime.datetime.utcnow()

            f = self.client.store_friend({
                **data,
                'favorite': _payload['favorite'],
                'direction': _payload['direction'],
                'status': _status,
                'created': timestamp,
            })

            try:
                del self.client._pending_friends[f.id]
            except KeyError:
                pass

            # Send presence to the newly added friend as that is now
            # required to do by the server (or at least thats what
            # i suspect)
            self.client.loop.create_task(self.send_presence(
                to=f.jid,
                status=self.client.party.last_raw_status,
                show=self.client.away.value
            ))

            self.client.dispatch_event('friend_add', f)

        elif _status == 'PENDING':
            data = self.client.get_user(_id)
            if data is None:
                data = await self.client.fetch_user(_id, raw=True)
            else:
                data = data.get_raw()

            data = {
                **data,
                'direction': _payload['direction'],
                'status': _status,
                'created': body['timestamp']
            }
            if _payload['direction'] == 'INBOUND':
                pf = self.client.store_incoming_pending_friend(data)
            else:
                pf = self.client.store_outgoing_pending_friend(data)

            self.client.dispatch_event('friend_request', pf)

    @dispatcher.event('FRIENDSHIP_REMOVE')
    async def friend_remove_event(self, ctx: EventContext) -> None:
        body = ctx.body

        if body['from'] == self.client.user.id:
            _id = body['to']
        else:
            _id = body['from']

        if body['reason'] == 'ABORTED':
            pf = self.client.get_pending_friend(_id)
            if pf is not None:
                self.client.store_user(pf.get_raw())

                try:
                    del self.client._pending_friends[pf.id]
                except KeyError:
                    pass

                self.client.dispatch_event('friend_request_abort', pf)
        elif body['reason'] == 'REJECTED':
            pf = self.client.get_pending_friend(_id)
            if pf is not None:
                self.client.store_user(pf.get_raw())

                try:
                    del self.client._pending_friends[pf.id]
                except KeyError:
                    pass

                self.client.dispatch_event('friend_request_decline', pf)
        else:
            f = self.client.get_friend(_id)
            if f is not None:
                self.client.store_user(f.get_raw())

                try:
                    del self.client._friends[f.id]
                except KeyError:
                    pass

                self.client.dispatch_event('friend_remove', f)

        try:
            del self.client._presences[_id]
        except KeyError:
            pass

    @dispatcher.event('com.epicgames.friends.core.apiobjects.BlockListEntryAdded')  # noqa
    async def event_blocklist_added(self, ctx: EventContext) -> None:
        body = ctx.body

        account_id = body['payload']['accountId']
        data = await self.client.fetch_user(account_id, raw=True)
        blocked_user = self.client.store_blocked_user(data)
        self.client.dispatch_event('user_block', blocked_user)

    @dispatcher.event('com.epicgames.friends.core.apiobjects.BlockListEntryRemoved')  # noqa
    async def event_blocklist_remove(self, ctx: EventContext) -> None:
        body = ctx.body

        account_id = body['payload']['accountId']
        user = await self.client.fetch_user(account_id)

        try:
            del self.client._blocked_users[user.id]
        except KeyError:
            pass

        self.client.dispatch_event('user_unblock', user)

    @dispatcher.event('com.epicgames.social.party.notification.v0.PING')
    async def event_ping_received(self, ctx: EventContext) -> None:
        body = ctx.body
        pinger = body['pinger_id']
        try:
            data = (await self.client.http.party_lookup_ping(pinger))[0]
        except (IndexError, HTTPException) as exc:
            if isinstance(exc, HTTPException):
                m = 'errors.com.epicgames.social.party.ping_not_found'
                if exc.message_code != m:
                    raise

            self.client.dispatch_event(
                'invalid_party_invite',
                self.client.get_friend(pinger)
            )
            return

        for inv in data['invites']:
            if inv['sent_by'] == pinger and inv['status'] == 'SENT':
                invite = inv
                break
        else:
            invite = self._create_invite(pinger, {**body, **data})

        if 'urn:epic:cfg:build-id_s' not in invite['meta']:
            pres = self.client.get_presence(pinger)
            if (pres is not None and pres.party is not None
                    and not pres.party.private):
                net_cl = pres.party.net_cl
            else:
                net_cl = self.client.net_cl
        else:
            s = invite['meta']['urn:epic:cfg:build-id_s']
            net_cl = s[4:] if s.startswith('1:1:') else s

        if net_cl != self.client.net_cl and self.client.net_cl != '':
            log.debug(
                'Could not match the currently set net_cl ({0!r}) to the '
                'received value ({1!r})'.format(self.client.net_cl, net_cl)
            )

        new_party = Party(self.client, data)
        await new_party._update_members(members=data['members'])

        invitation = ReceivedPartyInvitation(
            self.client,
            new_party,
            net_cl,
            invite
        )
        self.client.dispatch_event('party_invite', invitation)

    @dispatcher.event('com.epicgames.social.party.notification.v0.MEMBER_JOINED')  # noqa
    async def event_party_member_joined(self,
                                        ctx: EventContext) -> None:
        body = ctx.body

        user_id = body.get('account_id')
        if user_id != self.client.user.id:
            await self.client.wait_until_party_ready()

        party = self.client.party

        if party is None:
            return

        if party.id != body.get('party_id'):
            return

        if user_id == self.client.user.id:
            await self.client._internal_join_party_lock.wait()

        member = party.get_member(user_id)
        if member is None:
            member = (await party._update_members(
                (body,),
                remove_missing=False
            ))[0]

        if party.me is not None:
            party.me.do_on_member_join_patch()

            yielding = party.me._default_config.yield_leadership
            if party.me.leader and not yielding:
                fut = asyncio.ensure_future(party.refresh_squad_assignments())

        try:
            if member.id == self.client.user.id:
                await self.client.wait_for('muc_enter', timeout=2)
            else:
                def check(m):
                    return m.direct_jid.localpart == member.id

                await self.client.wait_for('muc_member_join',
                                           check=check,
                                           timeout=2)
        except asyncio.TimeoutError:
            pass

        try:
            await fut
        except UnboundLocalError:
            pass

        self.client.dispatch_event('party_member_join', member)

    @dispatcher.event('com.epicgames.social.party.notification.v0.MEMBER_LEFT')
    async def event_party_member_left(self, ctx: EventContext) -> None:
        body = ctx.body

        user_id = body.get('account_id')
        if user_id != self.client.user.id:
            await self.client._join_party_lock.wait()

        party = self.client.party

        if party is None:
            return

        if party.id != body.get('party_id'):
            return

        member = party.get_member(user_id)
        if member is None:
            return

        party._remove_member(member.id)

        if party.me and party.me.leader and member.id != party.me.id:
            await party.refresh_squad_assignments()

        self.client.dispatch_event('party_member_leave', member)

    @dispatcher.event('com.epicgames.social.party.notification.v0.MEMBER_KICKED')  # noqa
    async def event_party_member_kicked(self, ctx: EventContext) -> None:
        body = ctx.body

        user_id = body.get('account_id')
        if user_id != self.client.user.id:
            await self.client._join_party_lock.wait()

        party = self.client.party

        if party is None:
            return

        if party.id != body.get('party_id'):
            return

        member = party.get_member(user_id)
        if member is None:
            return

        party._remove_member(member.id)

        if party.me and party.me.leader and member.id != party.me.id:
            await party.refresh_squad_assignments()

        if member.id == self.client.user.id:
            await self.leave_muc()
            p = await self.client._create_party()

            self.client.party = p

        self.client.dispatch_event('party_member_kick', member)

    @dispatcher.event('com.epicgames.social.party.notification.v0.MEMBER_DISCONNECTED')  # noqa
    async def event_party_member_disconnected(self, ctx: EventContext) -> None:
        body = ctx.body
        user_id = body.get('account_id')

        if user_id != self.client.user.id:
            await self.client._join_party_lock.wait()

        party = self.client.party

        if party is None:
            return

        if party.id != body.get('party_id'):
            return

        member = party.get_member(user_id)
        if member is None:
            return

        # Dont continue processing for old connections
        data = await self.client.http.party_lookup(party.id)
        for member_data in data['members']:
            if member_data['account_id'] == user_id:
                connections = member_data['connections']
                if len(connections) == 1:
                    break

                for connection in connections:
                    if 'disconnected_at' not in connection:
                        return

        member._update_connection(body.get('connection'))
        self.client.dispatch_event('party_member_zombie', member)

    @dispatcher.event('com.epicgames.social.party.notification.v0.MEMBER_EXPIRED')  # noqa
    async def event_party_member_expired(self, ctx: EventContext) -> None:
        body = ctx.body

        user_id = body.get('account_id')
        if user_id != self.client.user.id:
            await self.client._join_party_lock.wait()

        party = self.client.party

        if party is None:
            return

        if party.id != body.get('party_id'):
            return

        member = party.get_member(user_id)
        if member is None:
            return

        party._remove_member(member.id)

        if party.me and party.me.leader and member.id != party.me.id:
            await party.refresh_squad_assignments()

        if member.id == self.client.user.id:
            p = await self.client._create_party()
            self.client.party = p

        self.client.dispatch_event('party_member_expire', member)

    @dispatcher.event('com.epicgames.social.party.notification.v0.MEMBER_CONNECTED')  # noqa
    async def event_party_member_connected(self, ctx: EventContext) -> None:
        body = ctx.body

        user_id = body.get('account_id')
        if user_id != self.client.user.id:
            await self.client._join_party_lock.wait()

        party = self.client.party

        if party is None:
            return

        if party.id != body.get('party_id'):
            return

        member = party.get_member(user_id)
        if member is None:
            return

        member._update_connection(body.get('connection'))
        if member.id == self.client.user.id:
            party.update_presence()

        self.client.dispatch_event('party_member_reconnect', member)

    @dispatcher.event('com.epicgames.social.party.notification.v0.MEMBER_NEW_CAPTAIN')  # noqa
    async def event_party_new_captain(self, ctx: EventContext) -> None:
        body = ctx.body
        party = ctx.party

        user_id = body.get('account_id')
        if user_id != self.client.user.id:
            await self.client._join_party_lock.wait()

        party = self.client.party

        if party is None:
            return

        if party.id != body.get('party_id'):
            return

        member = party.get_member(user_id)
        if member is None:
            return

        old_leader = party.leader
        party._update_roles(member)

        party.update_presence()
        self.client.dispatch_event('party_member_promote', old_leader, member)

    @dispatcher.event('com.epicgames.social.party.notification.v0.PARTY_UPDATED')  # noqa
    async def event_party_updated(self, ctx: EventContext) -> None:
        body = ctx.body

        user_id = body.get('account_id')
        if user_id != self.client.user.id:
            await self.client._join_party_lock.wait()

        party = self.client.party

        if party is None:
            return

        if party.id != body.get('party_id'):
            return

        def _getattr(member, key):
            value = getattr(member, key)
            if callable(value):
                value = value()
            return value

        check = {'playlist_info': 'playlist', 'squad_fill': None,
                 'privacy': None}
        pre_values = {k: _getattr(party, k) for k in check.keys()}

        party._update(body)
        self.client.dispatch_event('party_update', party)

        for key, pre_value in pre_values.items():
            value = _getattr(party, key)
            if pre_value != value:
                self.client.dispatch_event(
                    'party_{0}_change'.format(check[key] or key),
                    party,
                    pre_value,
                    value
                )

    @dispatcher.event('com.epicgames.social.party.notification.v0.MEMBER_STATE_UPDATED')  # noqa
    async def event_party_member_state_updated(self,
                                               ctx: EventContext) -> None:
        body = ctx.body

        user_id = body.get('account_id')
        if user_id != self.client.user.id:
            await self.client._join_party_lock.wait()

        party = self.client.party

        if party is None:
            return

        if party.id != body.get('party_id'):
            return

        member = party.get_member(user_id)
        if member is None:
            def check(m):
                return m.id == user_id

            try:
                member = await self.client.wait_for(
                    'party_member_join',
                    check=check,
                    timeout=1
                )
            except asyncio.TimeoutError:
                party_data = await self.client.http.party_lookup(party.id)
                for m_data in party_data['members']:
                    if user_id == m_data['account_id']:
                        member = (await party._update_members(
                            (m_data,),
                            remove_missing=False
                        ))[0]
                        break
                else:
                    if user_id == self.client.user.id:
                        await party._leave()
                        p = await self.client._create_party()
                        self.client.party = p
                    return

                yielding = party.me._default_config.yield_leadership
                if party.me and party.me.leader and not yielding:
                    await party.refresh_squad_assignments()

        def _getattr(member, key):
            value = getattr(member, key)
            if callable(value):
                value = value()
            return value

        check = ('ready', 'input', 'assisted_challenge', 'outfit', 'backpack',
                 'pet', 'pickaxe', 'contrail', 'emote', 'emoji', 'banner',
                 'battlepass_info', 'in_match', 'match_players_left',
                 'enlightenments', 'corruption', 'outfit_variants',
                 'backpack_variants', 'pickaxe_variants', 'contrail_variants')
        pre_values = {k: _getattr(member, k) for k in check}

        member.update(body)

        if party._default_config.team_change_allowed or not party.me.leader:
            req_j = body['member_state_updated'].get(
                'Default:MemberSquadAssignmentRequest_j'
            )
            if req_j is not None:
                req = json.loads(req_j)['MemberSquadAssignmentRequest']
                version = req['version']
                if version != member._assignment_version:
                    member._assignment_version = version

                    swap_member_id = req['swapTargetMemberId']
                    if swap_member_id != 'INVALID':
                        new_positions = {
                            member.id: req['targetAbsoluteIdx'],
                            swap_member_id: req['startingAbsoluteIdx']
                        }
                        if party.me.leader:
                            await party.refresh_squad_assignments(
                                new_positions=new_positions
                            )

                        try:
                            self.client.dispatch_event(
                                'party_member_team_swap',
                                *[party._members[k] for k in new_positions]
                            )
                        except KeyError:
                            pass

        self.client.dispatch_event('party_member_update', member)

        def _dispatch(key, member, pre_value, value):
            self.client.dispatch_event(
                'party_member_{0}_change'.format(key),
                member,
                pre_value,
                value
            )

        def compare(a, b):
            def construct_set(v):
                return set(itertools.chain(
                    *[list(x.values()) if isinstance(x, dict) else (x,)
                      for x in v]
                ))

            if isinstance(a, (tuple, list)) and isinstance(b, (tuple, list)):
                return construct_set(a) == construct_set(b)
            return a == b

        for key, pre_value in pre_values.items():
            value = _getattr(member, key)
            if not compare(pre_value, value):
                _dispatch(key, member, pre_value, value)

    @dispatcher.event('com.epicgames.social.party.notification.v0.MEMBER_REQUIRE_CONFIRMATION')  # noqa
    async def event_party_member_require_confirmation(self,
                                                      ctx: EventContext
                                                      ) -> None:
        body = ctx.body

        user_id = body.get('account_id')
        if user_id != self.client.user.id:
            await self.client._join_party_lock.wait()

        party = self.client.party

        if party is None:
            return

        if party.id != body.get('party_id'):
            return

        user = self.client.get_user(user_id)
        if user is None:
            user = await self.client.fetch_user(user_id)

        confirmation = PartyJoinConfirmation(self.client, party, user, body)

        # Automatically confirm if event is received but no handler is found.
        if not self.client._event_has_destination('party_member_confirm'):
            return await confirmation.confirm()

        self.client.dispatch_event('party_member_confirm', confirmation)

    @dispatcher.event('com.epicgames.social.party.notification.v0.INVITE_DECLINED')  # noqa
    async def event_party_invite_declined(self, ctx: EventContext) -> None:
        body = ctx.body

        friend = self.client.get_friend(body['invitee_id'])
        if friend is not None:
            self.client.dispatch_event('party_invite_decline', friend)

    @dispatcher.presence()
    async def process_presence(self, user_id: str,
                               platform: str,
                               type_: str,
                               status: str,
                               show: str) -> None:
        try:
            data = json.loads(status)

            # We do this to filter out kairos from launcher presences
            ch = data.get('bIsEmbedded', False) or data.get('Status', '') != ''

            is_dict = isinstance(data.get('Properties', {}), dict)
            if (not ch or 'bIsPlaying' not in data or not is_dict):
                return
        except ValueError:
            return

        friend = self.client.get_friend(user_id)
        if friend is None:
            try:
                friend = await self.client.wait_for(
                    'friend_add',
                    check=lambda f: f.id == user_id,
                    timeout=1
                )
            except asyncio.TimeoutError:
                return

        is_available = type_ is None or type_ == 'available'

        try:
            away = AwayStatus(show)
        except ValueError:
            away = AwayStatus.ONLINE

        _pres = Presence(
            self.client,
            user_id,
            platform,
            is_available,
            away,
            data
        )

        if _pres.party is not None:
            try:
                display_name = _pres.party.raw['sourceDisplayName']
                if display_name != _pres.friend.display_name:
                    _pres.friend._update_display_name(display_name)
            except (KeyError, AttributeError):
                pass

        before_pres = friend.last_presence

        if not is_available and friend.is_online():
            friend._update_last_logout(datetime.datetime.utcnow())

            try:
                del self.client._presences[user_id]
            except KeyError:
                pass

        else:
            self.client._presences[user_id] = _pres

        self.client.dispatch_event('friend_presence', before_pres, _pres)

    def on_stream_established(self):
        self.client.dispatch_event('xmpp_session_establish')

        async def on_establish():
            # Just incase the recover task hangs we don't want it
            # running forever in the background until a new close is
            # dispatched potentially fucking shit up big time.
            task = self._reconnect_recover_task
            if task is not None and not task.done():
                try:
                    await asyncio.wait_for(task, timeout=20)
                except asyncio.TimeoutError:
                    pass

        if self._is_suspended:
            self.client.dispatch_event('xmpp_session_reconnect')
            self.client.loop.create_task(self.client._reconnect_to_party())

        self._is_suspended = False
        self.client.loop.create_task(on_establish())

    def on_stream_suspended(self, reason):
        jid = self.xmpp_client.local_jid
        resource = jid.resource[:-32] + (uuid.uuid4().hex).upper()
        self.xmpp_client._local_jid = jid.replace(resource=resource)

        def on_events_recovered(*args):
            self._reconnect_recover_task = None

        self._reconnect_recover_task = task = self.client.loop.create_task(
            self.client.recover_events(
                refresh_caches=True,
                wait_for_close=False
            )
        )
        task.add_done_callback(on_events_recovered)

        if self.client.party is not None:
            self._last_known_party_id = self.client.party.id

        self._is_suspended = True
        self.client.dispatch_event('xmpp_session_lost')

    def on_stream_destroyed(self, reason=None):
        if not self._is_suspended:
            task = self._reconnect_recover_task
            if task is not None and not task.cancelled():
                task.cancel()

        self._last_disconnected_at = datetime.datetime.utcnow()
        self.client.dispatch_event('xmpp_session_close')

    def setup_callbacks(self, messages: bool = True) -> None:
        client = self.xmpp_client

        message_dispatcher = self.xmpp_client.summon(
            aioxmpp.dispatcher.SimpleMessageDispatcher
        )

        if messages:
            message_dispatcher.register_callback(
                aioxmpp.MessageType.CHAT,
                None,
                lambda m: asyncio.ensure_future(
                    self.process_chat_message(m)
                ),
            )

        client.on_stream_established.connect(self.on_stream_established)
        client.on_stream_suspended.connect(self.on_stream_suspended)
        client.on_stream_destroyed.connect(self.on_stream_destroyed)

    async def loop_ping(self) -> None:
        while True:
            await asyncio.sleep(60)
            iq = aioxmpp.IQ(
                type_=aioxmpp.IQType.GET,
                payload=aioxmpp.ping.Ping(),
                to=None,
            )
            await self.stream.send(iq)

    async def _run(self, future: asyncio.Future) -> None:
        async with self.xmpp_client.connected() as stream:
            self.stream = stream
            stream.soft_timeout = datetime.timedelta(minutes=3)
            stream.round_trip_time = datetime.timedelta(minutes=3)
            future.set_result(None)
            while True:
                await asyncio.sleep(1)

    async def run(self) -> None:
        resource_id = (uuid.uuid4().hex).upper()
        resource = 'V2:Fortnite:{0.client.platform.value}::{1}'.format(
            self,
            resource_id
        )

        self.xmpp_client = aioxmpp.PresenceManagedClient(
            aioxmpp.JID(
                self.client.user.id,
                self.client.service_host,
                resource
            ),
            aioxmpp.make_security_layer(
                self.client.auth.access_token,
                no_verify=True
            ),
            override_peer=[(
                self.client.service_domain,
                self.client.service_port,
                XMPPOverWebsocketConnector(
                    self.client,
                    ws_connector=self.ws_connector
                )
            )],
            loop=self.client.loop
        )
        self.xmpp_client.backoff_start = datetime.timedelta(seconds=0.1)

        self.muc_service = self.xmpp_client.summon(aioxmpp.MUCClient)
        self.setup_callbacks()

        future = self.client.loop.create_future()
        self._task = asyncio.ensure_future(self._run(future))
        await future

        self._ping_task = asyncio.ensure_future(self.loop_ping())

    async def close(self) -> None:
        log.debug('Attempting to close xmpp client')
        if self.xmpp_client.running:
            self.xmpp_client.stop()

            while self.xmpp_client.running:
                await asyncio.sleep(0)

        if self._task:
            self._task.cancel()
        if self._ping_task:
            self._ping_task.cancel()

        self._ping_task = None
        self.xmpp_client = None
        self.stream = None
        self.muc_service = None
        log.debug('Successfully closed xmpp client')

        # let loop run one iteration for events to be dispatched
        await asyncio.sleep(0)

    def muc_on_message(self, message: aioxmpp.Message,
                       member: aioxmpp.muc.Occupant,
                       source: aioxmpp.im.dispatcher.MessageSource,
                       **kwargs: Any) -> None:
        if member.direct_jid is None:
            return

        user_id = member.direct_jid.localpart
        party = self.client.party

        if (user_id == self.client.user.id or member.nick is None
                or user_id not in party._members):
            return

        self.client.dispatch_event('party_message', PartyMessage(
            client=self.client,
            party=party,
            author=party._members[member.direct_jid.localpart],
            content=message.body.any()
        ))

    def muc_on_member_join(self, member: aioxmpp.muc.Occupant) -> None:
        self.client.dispatch_event('muc_member_join', member)

    def muc_on_enter(self, *args: list, **kwargs: Any) -> None:
        self.client.dispatch_event('muc_enter')

    def muc_on_leave(self, member: aioxmpp.muc.Occupant,
                     muc_leave_mode: aioxmpp.muc.LeaveMode,
                     muc_actor: aioxmpp.muc.xso.UserActor,
                     muc_reason: str,
                     **kwargs: Any) -> None:
        if muc_leave_mode is aioxmpp.muc.LeaveMode.BANNED:
            mem = self.client.party._members[member.direct_jid.localpart]
            self.client.dispatch_event('party_member_chatban',
                                       mem,
                                       muc_reason)

    async def join_muc(self, party_id: str) -> None:
        muc_jid = aioxmpp.JID.fromstr(
            'Party-{}@muc.prod.ol.epicgames.com'.format(party_id)
        )
        nick = '{0}:{1}:{2}'.format(
            self._remove_illegal_characters(self.client.user.display_name),
            self.client.user.id,
            self.xmpp_client.local_jid.resource
        )

        room, fut = self.muc_service.join(
            muc_jid,
            nick
        )

        room.on_message.connect(self.muc_on_message)
        room.on_join.connect(self.muc_on_member_join)
        room.on_enter.connect(self.muc_on_enter)
        room.on_leave.connect(self.muc_on_leave)
        self.muc_room = room

        asyncio.ensure_future(fut)
        await self.client.wait_for('muc_enter')

    async def leave_muc(self) -> None:

        # come back to this. works sometimes? wait for 2 seconds and timeout
        # to manually leave + check if muc messages is sent from correct
        # room (party-id)

        if self.muc_room is not None:
            presence = aioxmpp.stanza.Presence(
                type_=aioxmpp.structs.PresenceType.UNAVAILABLE,
                to=self.muc_room._mucjid
            )
            await self.xmpp_client.send(presence)
            try:
                self.muc_service._muc_exited(self.muc_room)
            except KeyError:
                pass

    async def send_party_message(self, content: str) -> None:
        if self.muc_room is None:
            raise PartyError('Can\'t send message. Reason: No party found')

        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.GROUPCHAT
        )
        msg.body[None] = content
        self.muc_room.send_message(msg)

    async def send_friend_message(self, jid: aioxmpp.JID,
                                  content: str) -> None:
        if self.stream is None:
            raise XMPPError('xmpp is not connected')

        msg = aioxmpp.Message(
            to=jid,
            type_=aioxmpp.MessageType.CHAT,
        )
        msg.body[None] = content
        await self.stream.send(msg)

    def set_presence(self, *,
                     status: Optional[Union[str, dict]] = None,
                     show: Optional[str]) -> None:
        if status is None:
            self.xmpp_client.presence = aioxmpp.PresenceState(available=True)

        _status = status if isinstance(status, dict) else {'Status': status}
        self.xmpp_client.set_presence(
            state=aioxmpp.PresenceState(
                available=True,
                show=show
            ),
            status=json.dumps(_status)
        )

    async def send_presence(self, to: Optional[aioxmpp.JID] = None,
                            status: Optional[Union[str, dict]] = None,
                            show: Optional[str] = None) -> None:
        _status = {}
        if status is None:
            _status = None
        elif isinstance(status, str):
            _status['Status'] = status
        elif isinstance(status, dict):
            _status = status
        else:
            raise TypeError('status must be None, str or dict')

        pres = aioxmpp.Presence(
            type_=aioxmpp.PresenceType.AVAILABLE,
            show=aioxmpp.PresenceShow(show),
            to=to,
        )

        if _status is not None:
            pres.status[None] = json.dumps(_status)
        await self.stream.send(pres)

    async def get_presence(self, jid: aioxmpp.JID) -> Presence:
        self.client.loop.create_task(self.send_presence_probe(jid))
        _, after = await self.client.wait_for(
            'friend_presence',
            check=lambda b, a: a.friend.id == jid.localpart
        )
        return after

    async def send_presence_probe(self, to: aioxmpp.JID) -> None:
        presence = aioxmpp.Presence(
            type_=aioxmpp.PresenceType.PROBE,
            to=to
        )
        await self.stream.send(presence)



import re

from typing import Any, Optional
from vodajemokra.user import User
from vodajemokra.party import PartyMember
from vodajemokra.friend import Friend

from .errors import BadArgument
from .context import Context


__all__ = (
    'Converter',
    'IDConverter',
    'UserConverter',
    'PartyMemberConverter',
    'FriendConverter',
    'IDConverter',
    'Greedy',
)


class Converter:
    """The base class of custom converters that require the :class:`.Context`
    to be passed to be useful.

    This allows you to implement converters that function similar to the
    special cased ``vodajemokra`` classes.

    Classes that derive from this should override the
    :meth:`~.Converter.convert` method to do its conversion logic. This method
    must be a coroutine.
    """

    async def convert(self, ctx: 'Context', argument: str) -> Any:
        """|coro|

        The method to override to do conversion logic.

        If an error is found while converting, it is recommended to
        raise a :exc:`.CommandError` derived exception as it will
        properly propagate to the error handlers.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context that the argument is being used in.
        argument: :class:`str`
            The argument that is being converted.
        """
        raise NotImplementedError('Derived classes need to implement this.')


class IDConverter(Converter):
    def __init__(self) -> None:
        self._id_regex = re.compile(r'([0-9a-fA-F]{32})$')

    def _get_id_match(self, argument: str) -> Any:
        return self._id_regex.match(argument)


class UserConverter(IDConverter):
    """Converts to a :class:`~vodajemokra.User`.

    The lookup strategy is as follows (in order):
    1. Cache lookup by ID.
    2. Cache lookup by display name.
    3. API Request to fetch the user by id/display name.
    """

    async def convert(self, ctx: Context, argument: str) -> User:
        bot = ctx.bot
        match = self._get_id_match(argument)

        if match is not None:
            to_lookup = match.group(1).lower()
        else:
            to_lookup = argument

        result = await bot.fetch_user(to_lookup, cache=True)

        if result is None:
            raise BadArgument('User "{}" not found'.format(argument))

        return result


class PartyMemberConverter(IDConverter):
    """Converts to a :class:`~vodajemokra.PartyMember`.

    All lookups are done via the bots party member cache.

    The lookup strategy is as follows (in order):
    1. Lookup by ID.
    2. Lookup by display name.
    """

    async def convert(self, ctx: Context, argument: str) -> PartyMember:
        bot = ctx.bot
        match = self._get_id_match(argument)
        party = ctx.party or bot.party

        result = None
        if match is None:
            for member in party._members.values():
                if member.display_name.casefold() == argument.casefold():
                    result = member
                    break
        else:
            user_id = match.group(1).lower()
            result = party.get_member(user_id)

        if result is None:
            raise BadArgument('Party member "{}" not found'.format(argument))

        return result


class FriendConverter(IDConverter):
    """Converts to a :class:`~vodajemokra.Friend`.

    All lookups are via the friend cache.

    The lookup strategy is as follows (in order):
    1. Lookup by ID.
    2. Lookup by display name.
    """

    async def convert(self, ctx: Context, argument: str) -> Friend:
        bot = ctx.bot
        match = self._get_id_match(argument)

        result = None
        if match is not None:
            user_id = match.group(1).lower()
            result = bot.get_friend(user_id)
        else:
            for friend in bot._friends.values():
                if friend.display_name.casefold() == argument.casefold():
                    result = friend
                    break

        if result is None:
            raise BadArgument('Friend "{}" not found'.format(argument))

        return result


class _Greedy:
    __slots__ = ('converter',)

    def __init__(self, *, converter: Optional[Converter] = None) -> None:
        self.converter = converter

    def __getitem__(self, params: Any) -> Any:
        if not isinstance(params, tuple):
            params = (params,)
        if len(params) != 1:
            raise TypeError('Greedy[...] only takes a single argument')

        converter = params[0]

        if not (callable(converter) or isinstance(converter, Converter)
                or hasattr(converter, '__origin__')):
            raise TypeError('Greedy[...] expects a type or a Converter '
                            'instance.')

        NoneType = type(None)
        if converter is str or converter is NoneType or converter is _Greedy:
            raise TypeError('Greedy[%s] is invalid.' % converter.__name__)

        return self.__class__(converter=converter)


Greedy = _Greedy()



from inspect import Parameter
from typing import TYPE_CHECKING, Optional, List, Tuple
from vodajemokra.errors import FortniteException

if TYPE_CHECKING:
    from .cooldown import Cooldown, BucketType
    from .converter import Converter


__all__ = (
    'CommandError',
    'UserInputError',
    'CommandNotFound',
    'MissingRequiredArgument',
    'TooManyArguments',
    'BadArgument',
    'CheckFailure',
    'CheckAnyFailure',
    'PrivateMessageOnly',
    'PartyMessageOnly',
    'NotOwner',
    'DisabledCommand',
    'CommandInvokeError',
    'CommandOnCooldown',
    'MaxConcurrencyReached',
    'ConversionError',
    'BadUnionArgument',
    'ArgumentParsingError',
    'UnexpectedQuoteError',
    'InvalidEndOfQuotedStringError',
    'ExpectedClosingQuoteError',
    'ExtensionError',
    'ExtensionAlreadyLoaded',
    'ExtensionNotLoaded',
    'ExtensionMissingEntryPoint',
    'ExtensionFailed',
    'ExtensionNotFound',
)


class CommandError(FortniteException):
    r"""The base exception type for all command related errors.

    This inherits from :exc:`vodajemokra.FortniteException`.

    This exception and exceptions inherited from it are handled
    in a special way as they are caught and passed into a special event
    from :class:`.Bot`\, :func:`event_command_error`.
    """

    def __init__(self, message: Optional[str] = None, *args: list) -> None:
        if message is not None:
            super().__init__(message, *args)
        else:
            super().__init__(*args)


class UserInputError(CommandError):
    """The base exception type for errors that involve errors
    regarding user input.

    This inherits from :exc:`CommandError`.
    """
    pass


class CommandNotFound(CommandError):
    """Exception raised when a command is attempted to be invoked
    but no command under that name is found.

    This is not raised for invalid subcommands, rather just the
    initial main command that is attempted to be invoked.

    This inherits from :exc:`CommandError`.
    """
    pass


class MissingRequiredArgument(UserInputError):
    """Exception raised when parsing a command and a parameter
    that is required is not encountered.

    This inherits from :exc:`UserInputError`

    Attributes
    -----------
    param: :class:`inspect.Parameter`
        The argument that is missing.
    """

    def __init__(self, param: Parameter) -> None:
        self.param = param
        super().__init__('{0.name} is a required argument that is '
                         'missing.'.format(param))


class TooManyArguments(UserInputError):
    """Exception raised when the command was passed too many arguments and its
    :attr:`.Command.ignore_extra` attribute was not set to ``True``.

    This inherits from :exc:`UserInputError`
    """
    pass


class BadArgument(UserInputError):
    """Exception raised when a parsing or conversion failure is encountered
    on an argument to pass into a command.

    This inherits from :exc:`UserInputError`
    """
    pass


class CheckFailure(UserInputError):
    """Exception raised when the predicates in :attr:`.Command.checks` have failed.

    This inherits from :exc:`CommandError`
    """
    pass


class CheckAnyFailure(CheckFailure):
    """Exception raised when all predicates in :func:`check_any` fail.

    This inherits from :exc:`CheckFailure`.

    Attributes
    ------------
    errors: List[:class:`CheckFailure`]
        A list of errors that were caught during execution.
    checks: List[Callable[[:class:`Context`], :class:`bool`]]
        A list of check predicates that failed.
    """

    def __init__(self, checks: List[callable],
                 errors: List[CheckFailure]) -> None:
        self.checks = checks
        self.errors = errors
        super().__init__('You do not have permission to run this command.')


class PrivateMessageOnly(CheckFailure):
    """Exception raised when an operation does not work outside of private
    message contexts.

    This inherits from :exc:`CheckFailure`
    """

    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(
            message or 'This command can only be used in private messages.'
        )


class PartyMessageOnly(CheckFailure):
    """Exception raised when an operation does not work outside of party
    message contexts.

    This inherits from :exc:`CheckFailure`
    """
    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(
            message or 'This command can only be used in party messages.'
        )


class NotOwner(CheckFailure):
    """Exception raised when the message author is not the owner of the bot.

    This inherits from :exc:`CheckFailure`
    """
    pass


class DisabledCommand(CommandError):
    """Exception raised when the command being invoked is disabled.

    This inherits from :exc:`CommandError`
    """
    pass


class CommandInvokeError(CommandError):
    """Exception raised when the command being invoked raised an exception.

    This inherits from :exc:`CommandError`

    Attributes
    ----------
    original
        The original exception that was raised. You can also get this via
        the ``__cause__`` attribute.
    """

    def __init__(self, e: Exception) -> None:
        self.original = e
        super().__init__('Command raised an exception: '
                         '{0.__class__.__name__}: {0}'.format(e))


class CommandOnCooldown(CommandError):
    """Exception raised when the command being invoked is on cooldown.

    This inherits from :exc:`CommandError`

    Attributes
    -----------
    cooldown: Cooldown
        A class with attributes ``rate``, ``per``, and ``type`` similar to
        the :func:`.cooldown` decorator.
    retry_after: :class:`float`
        The amount of seconds to wait before you can retry again.
    """

    def __init__(self, cooldown: 'Cooldown', retry_after: float) -> None:
        self.cooldown = cooldown
        self.retry_after = retry_after
        super().__init__('You are on cooldown. Try again in '
                         '{:.2f}s'.format(retry_after))


class MaxConcurrencyReached(CommandError):
    """Exception raised when the command being invoked has reached its maximum
    concurrency.

    This inherits from :exc:`CommandError`.

    Attributes
    ----------
    number: :class:`int`
        The maximum number of concurrent invokers allowed.
    per: :class:`BucketType`
        The bucket type passed to the :func:`.max_concurrency` decorator.
    """

    def __init__(self, number: int, per: 'BucketType') -> None:
        self.number = number
        self.per = per
        name = per.name
        suffix = 'per %s' % name if per.name != 'default' else 'globally'
        plural = '%s times %s' if number > 1 else '%s time %s'
        fmt = plural % (number, suffix)
        super().__init__('Too many people using this command. It can only '
                         'be used {} concurrently.'.format(fmt))


class ConversionError(CommandError):
    """Exception raised when a Converter class raises non-CommandError.

    This inherits from :exc:`CommandError`.

    Attributes
    ----------
    converter: :class:`vodajemokra.ext.commands.Converter`
        The converter that failed.
    original
        The original exception that was raised. You can also get this via
        the ``__cause__`` attribute.
    """

    def __init__(self, converter: 'Converter', original: Exception) -> None:
        self.converter = converter
        self.original = original


class BadUnionArgument(UserInputError):
    """Exception raised when a :data:`typing.Union` converter fails for all
    its associated types.

    This inherits from :exc:`UserInputError`

    Attributes
    -----------
    param: :class:`inspect.Parameter`
        The parameter that failed being converted.
    converters: Tuple[Type, ...]
        A tuple of converters attempted in conversion, in order of failure.
    errors: List[:class:`CommandError`]
        A list of errors that were caught from failing the conversion.
    """

    def __init__(self, param: Parameter,
                 converters: Tuple['Converter'],
                 errors: List[CommandError]) -> None:
        self.param = param
        self.converters = converters
        self.errors = errors

        def _get_name(x):
            try:
                return x.__name__
            except AttributeError:
                return x.__class__.__name__

        to_string = [_get_name(x) for x in converters]
        if len(to_string) > 2:
            fmt = '{}, or {}'.format(', '.join(to_string[:-1]), to_string[-1])
        else:
            fmt = ' or '.join(to_string)

        super().__init__('Could not convert "{0.name}" into '
                         '{1}.'.format(param, fmt))


class ArgumentParsingError(UserInputError):
    """An exception raised when the parser fails to parse a user's input.

    This inherits from :exc:`UserInputError`.

    There are child classes that implement more granular parsing errors for
    i18n purposes.
    """
    pass


class UnexpectedQuoteError(ArgumentParsingError):
    """An exception raised when the parser encounters a quote mark inside a
    non-quoted string.

    This inherits from :exc:`ArgumentParsingError`.

    Attributes
    ------------
    quote: :class:`str`
        The quote mark that was found inside the non-quoted string.
    """

    def __init__(self, quote: str) -> None:
        self.quote = quote
        super().__init__('Unexpected quote mark, {0!r}, in non-quoted '
                         'string'.format(quote))


class InvalidEndOfQuotedStringError(ArgumentParsingError):
    """An exception raised when a space is expected after the closing quote in
    a string but a different character is found.

    This inherits from :exc:`ArgumentParsingError`.

    Attributes
    -----------
    char: :class:`str`
        The character found instead of the expected string.
    """

    def __init__(self, char: str) -> None:
        self.char = char
        super().__init__('Expected space after closing quotation but '
                         'received {0!r}'.format(char))


class ExpectedClosingQuoteError(ArgumentParsingError):
    """An exception raised when a quote character is expected but not found.

    This inherits from :exc:`ArgumentParsingError`.

    Attributes
    -----------
    close_quote: :class:`str`
        The quote character expected.
    """

    def __init__(self, close_quote: str) -> None:
        self.close_quote = close_quote
        super().__init__('Expected closing {}.'.format(close_quote))


# Extension

class ExtensionError(FortniteException):
    """Base exception for extension related errors.

    This inherits from :exc:`~vodajemokra.FortniteException`.

    Attributes
    ------------
    name: :class:`str`
        The extension that had an error.
    """

    def __init__(self, message: Optional[str] = None,
                 *args: list,
                 name: str) -> None:
        self.name = name
        message = message or 'Extension {!r} had an error.'.format(name)
        super().__init__(message, *args)


class ExtensionAlreadyLoaded(ExtensionError):
    """An exception raised when an extension has already been loaded.

    This inherits from :exc:`ExtensionError`
    """

    def __init__(self, name: str) -> None:
        super().__init__('Extension {!r} is already loaded.'
                         ''.format(name), name=name)


class ExtensionNotLoaded(ExtensionError):
    """An exception raised when an extension was not loaded.

    This inherits from :exc:`ExtensionError`
    """

    def __init__(self, name: str) -> None:
        super().__init__('Extension {!r} has not been loaded.'
                         ''.format(name), name=name)


class ExtensionMissingEntryPoint(ExtensionError):
    """An exception raised when an extension does not have a
    ``extension_setup`` entry point function.

    This inherits from :exc:`ExtensionError`
    """

    def __init__(self, name: str) -> None:
        super().__init__("Extension {!r} has no 'setup' function."
                         "".format(name), name=name)


class ExtensionFailed(ExtensionError):
    """An exception raised when an extension failed to load during execution
    of the module or ``extension_setup`` entry point.

    This inherits from :exc:`ExtensionError`

    Attributes
    -----------
    name: :class:`str`
        The extension that had the error.
    original: :exc:`Exception`
        The original exception that was raised. You can also get this via
        the ``__cause__`` attribute.
    """

    def __init__(self, name: str, original: Exception) -> None:
        self.original = original
        fmt = 'Extension {0!r} raised an error: {1.__class__.__name__}: {1}'
        super().__init__(fmt.format(name, original), name=name)


class ExtensionNotFound(ExtensionError):
    """An exception raised when an extension is not found.

    This inherits from :exc:`ExtensionError`

    Attributes
    -----------
    name: :class:`str`
        The extension that had the error.
    """

    def __init__(self, name: str) -> None:
        fmt = 'Extension {0!r} could not be loaded.'
        super().__init__(fmt.format(name), name=name)



# -*- coding: utf-8 -*-
# flake8: noqa

"""
MIT License

Copyright (c) 2019-2020 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import random
import types

from typing import Optional, Any
from enum import Enum as OriginalEnum


class Enum(OriginalEnum):
    @classmethod
    def get_random_member(cls) -> Optional[Any]:
        try:
            return cls[random.choice(cls._member_names_)]
        except IndexError:
            pass

    @classmethod
    def get_random_name(cls) -> Optional[Any]:
        member = cls.get_random_member()
        if member is not None:
            return member.name

    @classmethod
    def get_random_value(cls) -> Optional[Any]:
        member = cls.get_random_member()
        if member is not None:
            return member.value


class PartyPrivacy(Enum):
    PUBLIC = {
        'partyType': 'Public',
        'inviteRestriction': 'AnyMember',
        'onlyLeaderFriendsCanJoin': False,
        'presencePermission': 'Anyone',
        'invitePermission': 'Anyone',
        'acceptingMembers': True,
    }
    FRIENDS_ALLOW_FRIENDS_OF_FRIENDS = {
        'partyType': 'FriendsOnly',
        'inviteRestriction': 'AnyMember',
        'onlyLeaderFriendsCanJoin': False,
        'presencePermission': 'Anyone',
        'invitePermission': 'AnyMember',
        'acceptingMembers': True,
    }
    FRIENDS = {
        'partyType': 'FriendsOnly',
        'inviteRestriction': 'LeaderOnly',
        'onlyLeaderFriendsCanJoin': True,
        'presencePermission': 'Leader',
        'invitePermission': 'Leader',
        'acceptingMembers': False,
    }
    PRIVATE_ALLOW_FRIENDS_OF_FRIENDS = {
        'partyType': 'Private',
        'inviteRestriction': 'AnyMember',
        'onlyLeaderFriendsCanJoin': False,
        'presencePermission': 'Noone',
        'invitePermission': 'AnyMember',
        'acceptingMembers': False,
    }
    PRIVATE = {
        'partyType': 'Private',
        'inviteRestriction': 'LeaderOnly',
        'onlyLeaderFriendsCanJoin': True,
        'presencePermission': 'Noone',
        'invitePermission': 'Leader',
        'acceptingMembers': False,
    }


class PartyDiscoverability(Enum):
    ALL          = 'ALL'
    INVITED_ONLY = 'INVITED_ONLY'


class PartyJoinability(Enum):
    OPEN              = 'OPEN'
    INVITE_ONLY       = 'INVITE_ONLY'
    INVITE_AND_FORMER = 'INVITE_AND_FORMER' 


class DefaultCharactersChapter1(Enum):
    CID_001_Athena_Commando_F_Default = 1
    CID_002_Athena_Commando_F_Default = 2
    CID_003_Athena_Commando_F_Default = 3
    CID_004_Athena_Commando_F_Default = 4
    CID_005_Athena_Commando_M_Default = 5
    CID_006_Athena_Commando_M_Default = 6
    CID_007_Athena_Commando_M_Default = 7
    CID_008_Athena_Commando_M_Default = 8


class DefaultCharactersChapter2(Enum):
    CID_556_Athena_Commando_F_RebirthDefaultA = 1
    CID_557_Athena_Commando_F_RebirthDefaultB = 2
    CID_558_Athena_Commando_F_RebirthDefaultC = 3
    CID_559_Athena_Commando_F_RebirthDefaultD = 4
    CID_560_Athena_Commando_M_RebirthDefaultA = 5
    CID_561_Athena_Commando_M_RebirthDefaultB = 6
    CID_562_Athena_Commando_M_RebirthDefaultC = 7
    CID_563_Athena_Commando_M_RebirthDefaultD = 8


class V1Gamemode(Enum):
    SOLO  = 'p2'
    DUO   = 'p10'
    SQUAD = 'p9'


class V1Platform(Enum):
    PC   = 'pc'
    XBOX = 'xb1'
    PS4  = 'ps4'


class V1Window(Enum):
    ALLTIME = 'alltime'
    WEEKLY  = 'weekly'


class V2Input(Enum):
    KEYBOARDANDMOUSE = 'keyboardmouse'
    GAMEPAD          = 'gamepad'
    TOUCH            = 'touch'


class Region(Enum):
    NAEAST     = 'NAE'
    NAWEST     = 'NAW'
    EUROPE     = 'EU'
    BRAZIL     = 'BR'
    OCEANIA    = 'OCE'
    ASIA       = 'ASIA'
    MIDDLEEAST = 'ME'


class Platform(Enum):
    WINDOWS     = 'WIN'
    MAC         = 'MAC'
    PLAYSTATION = 'PSN'
    XBOX        = 'XBL'
    SWITCH      = 'SWT'
    IOS         = 'IOS'
    ANDROID     = 'AND'


class UserSearchPlatform(Enum):
    EPIC_GAMES  = 'epic'
    PLAYSTATION = 'psn'
    XBOX        = 'xbl'


class UserSearchMatchType(Enum):
    EXACT = 'exact'
    PREFIX = 'prefix'


class ReadyState(Enum):
    READY       = 'Ready'
    NOT_READY   = 'NotReady'
    SITTING_OUT = 'SittingOut'


class AwayStatus(Enum):
    ONLINE        = None
    AWAY          = 'away'
    EXTENDED_AWAY = 'xa'


class SeasonStartTimestamp(Enum):
    SEASON_1  = 1508889601
    SEASON_2  = 1513209601
    SEASON_3  = 1519257601
    SEASON_4  = 1525132801
    SEASON_5  = 1531353601
    SEASON_6  = 1538006401
    SEASON_7  = 1544054401
    SEASON_8  = 1551312001
    SEASON_9  = 1557360001
    SEASON_10 = 1564617601
    SEASON_11 = 1571097601
    SEASON_12 = 1582156801
    SEASON_13 = 1592352001
    SEASON_14 = 1598486401


class SeasonEndTimestamp(Enum):
    SEASON_1  = 1513123200
    SEASON_2  = 1519171200
    SEASON_3  = 1525046400
    SEASON_4  = 1531353600
    SEASON_5  = 1538006400
    SEASON_6  = 1544054400
    SEASON_7  = 1551312000
    SEASON_8  = 1557360000
    SEASON_9  = 1564617600
    SEASON_10 = 1570924800
    SEASON_11 = 1582156800
    SEASON_12 = 1592352000
    SEASON_13 = 1598486400


class BattlePassStat(Enum):
    SEASON_11 = ('s11_social_bp_level', SeasonEndTimestamp.SEASON_11.value)
    SEASON_12 = ('s11_social_bp_level', SeasonEndTimestamp.SEASON_12.value)
    SEASON_13 = (('s13_social_bp_level', 's11_social_bp_level'), SeasonEndTimestamp.SEASON_13.value)
    SEASON_14 = ('s14_social_bp_level', None)


class KairosBackgroundColorPreset(Enum):
    TEAL         = ["#8EFDE5","#1CBA9E","#034D3F"]
    SWEET_RED    = ["#FF81AE","#D8033C","#790625"]
    LIGHT_ORANGE = ["#FFDF00","#FBA000","#975B04"]
    GREEN        = ["#CCF95A","#30C11B","#194D12"]
    LIGHT_BLUE   = ["#B4F2FE","#00ACF2","#005679"]
    DARK_BLUE    = ["#1CA2E6","#0C5498","#081E3E"]
    PINK         = ["#FFB4D6","#FF619C","#7D3449"]
    RED          = ["#F16712","#D8033C","#6E0404"]
    GRAY         = ["#AEC1D3","#687B8E","#36404A"]
    ORANGE       = ["#FFAF5D","#FF6D32","#852A05"]
    DARK_PURPLE  = ["#E93FEB","#7B009C","#500066"]
    LIME         = ["#DFFF73","#86CF13","#404B07"]
    INDIGO       = ["#B35EEF","#4D1397","#2E0A5D"]


class StatsCollectionType(Enum):
    FISH = 'collection_fish'



import inspect
import asyncio

from typing import (TYPE_CHECKING, Union, Awaitable, Any, List, Dict, Iterable,
                    Optional)
from vodajemokra.typedefs import MaybeCoro

from ._types import _BaseCommand
from .context import Context

if TYPE_CHECKING:
    from .bot import Bot


__all__ = (
    'CogMeta',
    'Cog',
)


class CogMeta(type):
    """A metaclass for defining a cog.

    Note that you should probably not use this directly. It is exposed
    purely for documentation purposes.

    .. note::

        When passing an attribute of a metaclass that is documented below, note
        that you must pass it as a keyword-only argument to the class creation
        like the following example:

        .. code-block:: python3

            class MyCog(commands.Cog, name='My Cog'):
                pass

    Attributes
    -----------
    name: :class:`str`
        The cog name. By default, it is the name of the class with no
        modification.
    command_attrs: :class:`dict`
        A list of attributes to apply to every command inside this cog.
        The dictionary is passed into the :class:`Command` (or its subclass)
        options at ``__init__``. If you specify attributes inside the command
        attribute in the class, it will override the one specified inside this
        attribute. For example:

        .. code-block:: python3

            class MyCog(commands.Cog, command_attrs=dict(hidden=True)):

                @commands.command()
                async def foo(self, ctx):
                    pass # hidden -> True

                @commands.command(hidden=False)
                async def bar(self, ctx):
                    pass # hidden -> False
    """

    def __new__(cls, *args: list, **kwargs: dict) -> None:
        name, bases, attrs = args
        attrs['__cog_name__'] = cog_name = kwargs.pop('name', name)

        last = cog_name.split()[-1]
        try:
            int(last)
        except ValueError:
            pass
        else:
            raise NameError('Cog names cannot end with a space followed by a '
                            'number or solely consist of a number.')

        command_attrs = kwargs.pop('command_attrs', {})
        attrs['__cog_settings__'] = command_attrs

        commands = {}
        event_handlers = {}
        no_bot_cog = ('Commands or listeners must not start with cog_ or bot_ '
                      '(in method {0.__name__}.{1})')

        new_cls = super().__new__(cls, name, bases, attrs, **kwargs)
        for base in reversed(new_cls.__mro__):
            for elem, value in base.__dict__.items():
                if elem in commands:
                    del commands[elem]
                if elem in event_handlers:
                    del event_handlers[elem]

                is_static_method = isinstance(value, staticmethod)
                if is_static_method:
                    value = value.__func__

                if isinstance(value, _BaseCommand):
                    if is_static_method:
                        raise TypeError(
                            'Command in method {0}.{1!r} must not be '
                            'staticmethod.'.format(base, elem)
                        )

                    commands[elem] = value

                elif asyncio.iscoroutinefunction(value):
                    try:
                        getattr(value, '__cog_event_handler__')
                    except AttributeError:
                        pass
                    else:
                        if elem.startswith(('cog_', 'bot_')):
                            raise TypeError(no_bot_cog.format(base, elem))
                        event_handlers[elem] = value

        new_cls.__cog_commands__ = list(commands.values())

        event_handlers_as_list = []
        for handler in event_handlers.values():
            for handler_name, was_def in handler.__cog_event_handler_names__:
                event_handlers_as_list.append((
                    handler_name,
                    handler.__name__,
                    was_def
                ))

        new_cls.__cog_event_handlers__ = event_handlers_as_list
        return new_cls

    def __init__(self, *args: list, **kwargs: dict) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def qualified_name(cls) -> str:
        return cls.__cog_name__


def _cog_special_method(func: MaybeCoro) -> MaybeCoro:
    func.__cog_special_method__ = None
    return func


class Cog(metaclass=CogMeta):
    """The base class that all cogs must inherit from.

    A cog is a collection of commands, listeners, and optional state to
    help group commands together. More information on them can be found on
    the :ref:`ext_commands_cogs` page.

    When inheriting from this class, the options shown in :class:`CogMeta`
    are equally valid here.
    """

    def __new__(cls, *args: list, **kwargs: dict) -> None:
        self = super().__new__(cls)
        cmd_attrs = cls.__cog_settings__

        self.__cog_commands__ = tuple(
            c._update_copy(cmd_attrs) for c in cls.__cog_commands__
        )

        lookup = {
            cmd.qualified_name: cmd for cmd in self.__cog_commands__
        }

        for command in self.__cog_commands__:
            setattr(self, command.callback.__name__, command)
            parent = command.parent
            if parent is not None:
                parent = lookup[parent.qualified_name]

                parent.remove_command(command.name)
                parent.add_command(command)

        return self

    def get_commands(self) -> List[_BaseCommand]:
        r"""Returns a :class:`list` of :class:`.Command`\s that are
        defined inside this cog.

        .. note::

            This does not include subcommands.
        """
        return [c for c in self.__cog_commands__ if c.parent is None]

    @property
    def qualified_name(self) -> str:
        """:class:`str`: Returns the cog's specified name, not
        the class name.
        """
        return self.__cog_name__

    @property
    def description(self) -> str:
        """:class:`str`: Returns the cog's description, typically
        the cleaned docstring.
        """
        try:
            return self.__cog_cleaned_doc__
        except AttributeError:
            self.__cog_cleaned_doc__ = inspect.getdoc(self)
            return self.__cog_cleaned_doc__

    def walk_commands(self) -> Iterable:
        """An iterator that recursively walks through this cog's
        commands and subcommands.
        """
        from .core import GroupMixin

        for command in self.__cog_commands__:
            if command.parent is None:
                yield command
                if isinstance(command, GroupMixin):
                    yield from command.walk_commands()

    @property
    def event_handlers(self) -> Dict[str, Awaitable]:
        """:class:`list`: A list of (name, function) event handler pairs
        that are defined in this cog.
        """
        return {
            name: getattr(self, method_name)
            for name, method_name, _
            in self.__cog_event_handlers__
        }

    @classmethod
    def _get_overridden_method(cls, method: MaybeCoro) -> MaybeCoro:
        """Return None if the method is not overridden. Otherwise returns the
        overridden method.
        """
        return getattr(method.__func__, '__cog_special_method__', method)

    @classmethod
    def event(cls, event: Union[str, Awaitable[Any]] = None) -> Awaitable:
        """A decorator to register an event.

        Usage: ::

            @commands.Cog.event()
            async def event_friend_message(message):
                await message.reply('Thanks for your message!')

            @commands.Cog.event('friend_message')
            async def my_message_handler(message):
                await message.reply('Thanks for your message!')

        Raises
        ------
        TypeError
            The decorated function is not a coroutine.
        TypeError
            Event is not specified as argument or function name with event
            prefix.
        """
        is_coro = callable(event)

        def pred(coro):
            if isinstance(coro, staticmethod):
                coro = coro.__func__

            if not asyncio.iscoroutinefunction(coro):
                raise TypeError('the decorated function must be a coroutine')

            if is_coro or event is None:
                name = coro.__name__
                name_defined = False
            else:
                name = event
                name_defined = True

            coro.__cog_event_handler__ = True

            try:
                coro.__cog_event_handler_names__.append((name, name_defined))
            except AttributeError:
                coro.__cog_event_handler_names__ = [(name, name_defined)]

            return coro
        return pred(event) if is_coro else pred

    def _inject(self, bot: 'Bot') -> 'Cog':
        cls = self.__class__

        for index, command in enumerate(self.__cog_commands__):
            command.cog = self
            command.instance = bot
            if command.parent is None:
                try:
                    bot.add_command(command)
                except Exception as e:
                    for to_undo in self.__cog_commands__[:index]:
                        bot.remove_command(to_undo)
                    raise e

        if cls.bot_check is not Cog.bot_check:
            bot.add_check(self.bot_check)

        if cls.bot_check_once is not Cog.bot_check_once:
            bot.add_check(self.bot_check_once, call_once=True)

        # We need to do this to remove all prefixes of events registered
        # with the coro's name instead of manually specifying the name.
        new = []
        for name, method, name_was_defined in self.__cog_event_handlers__:
            if not name_was_defined:
                if name.startswith(bot.event_prefix):
                    name = name[len(bot.event_prefix):]

            new.append((name, method, name_was_defined))

        self.__cog_event_handlers__ = new

        for name, method_name, _ in self.__cog_event_handlers__:
            bot.add_event_handler(name, getattr(self, method_name))

        return self

    def _eject(self, bot: 'Bot') -> None:
        cls = self.__class__

        try:
            for command in self.__cog_commands__:
                if command.parent is None:
                    bot.remove_command(command.name)

            for name, method_name, _ in self.__cog_event_handlers__:
                bot.remove_event_handler(name, getattr(self, method_name))

            if cls.bot_check is not Cog.bot_check:
                bot.remove_check(self.bot_check)

            if cls.bot_check_once is not Cog.bot_check_once:
                bot.remove_check(self.bot_check_once, call_once=True)

        finally:
            self.cog_unload()

    @_cog_special_method
    def cog_unload(self) -> None:
        """A special method that is called when the cog gets removed.

        This function **cannot** be a coroutine. It must be a regular
        function.

        Subclasses must replace this if they want special unloading behaviour.
        """
        pass

    @_cog_special_method
    def bot_check_once(self, ctx: 'Context') -> bool:
        """A special method that registers as a :meth:`.Bot.check_once`
        check.

        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def bot_check(self, ctx: 'Context') -> bool:
        """A special method that registers as a :meth:`.Bot.check`
        check.

        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def cog_check(self, ctx: 'Context') -> bool:
        """A special method that registers as a :func:`commands.check`
        for every command and subcommand in this cog.

        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def cog_command_error(self, ctx: 'Context',
                          error: Exception) -> Optional[bool]:
        """A special method that is called whenever an error
        is dispatched inside this cog.

        This is similar to :func:`.event_command_error` except only applying
        to the commands inside this cog.

        Command error handlers are raised in a specific order. Returning
        ``False`` in any of them will invoke the next handler in the chain. If
        there are no handlers left to call, the error is printed.

        The order goes as follows:
        1. The local command error handler is called. (Handler specified by
        :meth:`.Command.check`)
        2. The local cog command error handler is called. (This)
        3. All :func:`.event_command_error()` handlers are called
        simultaneously. If any of them return False, the error will
        be printed.

        This **must** be a coroutine.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context where the error happened.
        error: :class:`CommandError`
            The error that happened.
        """
        return False

    @_cog_special_method
    async def cog_before_invoke(self, ctx: 'Context') -> None:
        """A special method that acts as a cog local pre-invoke hook.

        This is similar to :meth:`.Command.before_invoke`.

        This **must** be a coroutine.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context.
        """
        pass

    @_cog_special_method
    async def cog_after_invoke(self, ctx: Context) -> None:
        """A special method that acts as a cog local post-invoke hook.

        This is similar to :meth:`.Command.after_invoke`.

        This **must** be a coroutine.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context.
        """
        pass



from datetime import datetime
from typing import Union, Awaitable, Callable


MaybeCoro = Union[Awaitable, Callable]
StrOrMaybeCoro = Union[str, MaybeCoro]
ListOrTuple = Union[list, tuple]
StrOrInt = Union[str, int]
DatetimeOrTimestamp = Union[datetime, int]



import time
import asyncio

from vodajemokra.enums import Enum
from collections import deque
from typing import Any, Optional

from .typedefs import Message
from .errors import MaxConcurrencyReached

__all__ = (
    'BucketType',
    'Cooldown',
    'CooldownMapping',
    'MaxConcurrency',
)


class BucketType(Enum):
    default = 0
    user = 1

    def get_key(self, msg: Message) -> Any:
        if self is BucketType.user:
            return msg.author.id


class Cooldown:
    __slots__ = ('rate', 'per', 'type', '_window', '_tokens', '_last')

    def __init__(self, rate: int, per: float, type: BucketType) -> None:
        self.rate = int(rate)
        self.per = float(per)
        self.type = type
        self._window = 0.0
        self._tokens = self.rate
        self._last = 0.0

        if not isinstance(self.type, BucketType):
            raise TypeError('Cooldown type must be a BucketType')

    def get_tokens(self, current: Optional[float] = None) -> int:
        if not current:
            current = time.time()

        tokens = self._tokens

        if current > self._window + self.per:
            tokens = self.rate
        return tokens

    def update_rate_limit(self,
                          current: Optional[float] = None) -> Optional[float]:
        current = current or time.time()
        self._last = current

        self._tokens = self.get_tokens(current)

        # first token used means that we start a new rate limit window
        if self._tokens == self.rate:
            self._window = current

        # check if we are rate limited
        if self._tokens == 0:
            return self.per - (current - self._window)

        # we're not so decrement our tokens
        self._tokens -= 1

        # see if we got rate limited due to this token change, and if
        # so update the window to point to our current time frame
        if self._tokens == 0:
            self._window = current

    def reset(self) -> None:
        self._tokens = self.rate
        self._last = 0.0

    def copy(self) -> 'Cooldown':
        return Cooldown(self.rate, self.per, self.type)

    def __repr__(self) -> str:
        return ('<Cooldown rate: {0.rate} per: {0.per} window: {0._window} '
                'tokens: {0._tokens}>'.format(self))


class CooldownMapping:
    def __init__(self, original: Cooldown) -> None:
        self._cache = {}
        self._cooldown = original

    def copy(self) -> 'CooldownMapping':
        ret = CooldownMapping(self._cooldown)
        ret._cache = self._cache.copy()
        return ret

    @property
    def valid(self) -> bool:
        return self._cooldown is not None

    @classmethod
    def from_cooldown(cls, rate: int,
                      per: float,
                      type: BucketType) -> 'CooldownMapping':
        return cls(Cooldown(rate, per, type))

    def _bucket_key(self, msg: Message) -> Any:
        return self._cooldown.type.get_key(msg)

    def _verify_cache_integrity(self, current: Optional[float] = None) -> None:
        # we want to delete all cache objects that haven't been used
        # in a cooldown window. e.g. if we have a  command that has a
        # cooldown of 60s and it has not been used in 60s then that key should
        # be deleted
        current = current or time.time()
        dead_keys = [k for k, v in self._cache.items()
                     if current > v._last + v.per]
        for k in dead_keys:
            del self._cache[k]

    def get_bucket(self, message: Message,
                   current: Optional[float] = None) -> Cooldown:
        if self._cooldown.type is BucketType.default:
            return self._cooldown

        self._verify_cache_integrity(current)
        key = self._bucket_key(message)
        if key not in self._cache:
            bucket = self._cooldown.copy()
            self._cache[key] = bucket
        else:
            bucket = self._cache[key]

        return bucket

    def update_rate_limit(self, message: Message,
                          current: Optional[float] = None) -> Optional[float]:
        bucket = self.get_bucket(message, current)
        return bucket.update_rate_limit(current)


class _Semaphore:
    __slots__ = ('value', 'loop', '_waiters')

    def __init__(self, number: int) -> None:
        self.value = number
        self.loop = asyncio.get_event_loop()
        self._waiters = deque()

    def __repr__(self) -> str:
        return ('<_Semaphore value={0.value} waiters={1}>'
                ''.format(self, len(self._waiters)))

    def locked(self) -> bool:
        return self.value == 0

    def is_active(self) -> bool:
        return len(self._waiters) > 0

    def wake_up(self) -> None:
        while self._waiters:
            future = self._waiters.popleft()
            if not future.done():
                future.set_result(None)
                return

    async def acquire(self, *, wait: bool = False) -> bool:
        if not wait and self.value <= 0:
            # signal that we're not acquiring
            return False

        while self.value <= 0:
            future = self.loop.create_future()
            self._waiters.append(future)
            try:
                await future
            except Exception:
                future.cancel()
                if self.value > 0 and not future.cancelled():
                    self.wake_up()
                raise

        self.value -= 1
        return True

    def release(self) -> None:
        self.value += 1
        self.wake_up()


class MaxConcurrency:
    __slots__ = ('number', 'per', 'wait', '_mapping')

    def __init__(self, number: int, *, per: BucketType, wait: bool) -> None:
        self._mapping = {}
        self.per = per
        self.number = number
        self.wait = wait

        if number <= 0:
            raise ValueError('max_concurrency \'number\' cannot be less '
                             'than 1')

        if not isinstance(per, BucketType):
            raise TypeError('max_concurrency \'per\' must be of type '
                            'BucketType not %r' % type(per))

    def copy(self) -> 'MaxConcurrency':
        return self.__class__(self.number, per=self.per, wait=self.wait)

    def __repr__(self) -> str:
        return ('<MaxConcurrency per={0.per!r} number={0.number} '
                'wait={0.wait}>'.format(self))

    def get_key(self, message: Message) -> Any:
        return self.per.get_key(message)

    async def acquire(self, message: Message) -> None:
        key = self.get_key(message)

        try:
            sem = self._mapping[key]
        except KeyError:
            self._mapping[key] = sem = _Semaphore(self.number)

        acquired = await sem.acquire(wait=self.wait)
        if not acquired:
            raise MaxConcurrencyReached(self.number, self.per)

    async def release(self, message: Message) -> None:
        # Technically there's no reason for this function to be async
        # But it might be more useful in the future
        key = self.get_key(message)

        try:
            sem = self._mapping[key]
        except KeyError:
            # ...? peculiar
            return
        else:
            sem.release()

        if sem.value >= self.number and not sem.is_active():
            del self._mapping[key]



import asyncio

from typing import Optional


class MaybeLock:
    def __init__(self, lock: asyncio.Lock,
                 loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        self.lock = lock
        self.loop = loop or asyncio.get_event_loop()
        self._cleanup = False

    async def _acquire(self) -> None:
        await self.lock.acquire()
        self._cleanup = True

    async def __aenter__(self) -> 'MaybeLock':
        self._task = self.loop.create_task(self._acquire())
        return self

    async def __aexit__(self, *args: list) -> None:
        if not self._task.cancelled():
            self._task.cancel()

        if self._cleanup:
            self.lock.release()


class LockEvent(asyncio.Lock):
    def __init__(self) -> None:
        super().__init__()

        self._event = asyncio.Event()
        self._event.set()
        self.wait = self._event.wait
        self.priority = 0

    async def acquire(self) -> None:
        await super().acquire()
        self._event.clear()

    def release(self) -> None:
        super().release()

        # Only set if no new acquire waiters exists. This is because we
        # don't want any wait()'s to return if there immediately will
        # be a new acquirer.
        if not (self._waiters is not None and [w for w in self._waiters
                                               if not w.cancelled()]):
            self._event.set()
            self.priority = 0
